local MAJOR = "LibUnitID-1.0"
--@alpha@
MAJOR = MAJOR.."-alpha"
--@end-alpha@
local MINOR = 9999
--@non-debug@
MINOR = 3
--@end-non-debug@
local LibUnitID,oldminor = LibStub:NewLibrary(MAJOR,MINOR)
if (not LibUnitID) then return end -- already loaded, no upgrade necessary

-- dump data from old version, if any
local oldframe,oldlisteners,oldcacheupdates
if (type(LibUnitID.__dump_for_upgrade) == "function") then
	oldframe,oldlisteners,oldcacheupdates = LibUnitID:__dump_for_upgrade()
end

-- localize upvalues
local next,pairs,pcall,wipe = next,pairs,pcall,wipe
local API_UnitGUID = UnitGUID

-- hash table: tags that can be tracked with events
local tagSupported = {
	--player not needed since it never changes
	vehicle=1,vehicletarget=1,
	pet=1,pettarget=1,
	target=1,targettarget=1,
	focus=1,focustarget=1
}
for i=1,5 do tagSupported["party"..i]=1; tagSupported["partypet"..i]=1; tagSupported["partypet"..i.."target"]=1; end
for i=1,40 do tagSupported["raid"..i]=1; tagSupported["raidpet"..i]=1; tagSupported["raidpet"..i.."target"]=1; end

-- hash table: tags that are the pet of (thus dependent on) other tags
local tagPet = {
	player="pet"
	--player="vehicle" is equally valid but doesn't fit the scheme very well, so there's a special check for that in UNIT_PET
	--target and focus get UNIT_PET events, but targetpet and focuspet aren't valid unitIDs.  boggle
}
for i=1,5 do tagPet["party"..i]="partypet"..i; end
for i=1,40 do tagPet["raid"..i]="raidpet"..i; end

-- hash table: tags that are the target of (thus dependent on) other tags
local tagTarget = {
	--player="target" not needed since we use PLAYER_TARGET_CHANGED instead of UNIT_TARGET(player)
	vehicle="vehicletarget",
	pet="pettarget",
	target="targettarget",
	focus="focustarget"
}
for i=1,5 do tagTarget["party"..i]="party"..i.."target"; tagTarget["partypet"..i]="partypet"..i.."target"; end
for i=1,40 do tagTarget["raid"..i]="raid"..i.."target"; tagTarget["raidpet"..i]="raidpet"..i.."target"; end

-- hash table: tags that depend on another tag
local tagParent = {
	--target="player" not needed since player never changes
	--vehicle="player", ditto
	vehicletarget="vehicle",
	pettarget="pet",
	targettarget="target",
	focustarget="focus"
}
for i=1,5 do tagParent["party"..i.."target"]="party"..i; tagParent["partypet"..i]="party"..i; tagParent["partypet"..i.."target"]="partypet"..i; end
for i=1,40 do tagParent["raid"..i.."target"]="raid"..i; tagParent["raidpet"..i]="raid"..i; tagParent["raidpet"..i.."target"]="raidpet"..i; end

-- hash table: tags that depend on the party composition
local partyTags = {}
for i=1,5 do partyTags["party"..i]=1; partyTags["party"..i.."target"]=1; partyTags["partypet"..i]=1; partyTags["partypet"..i.."target"]=1; end

-- hash table: tags that depend on the raid composition
local raidTags = {}
for i=1,40 do raidTags["raid"..i]=1; raidTags["raid"..i.."target"]=1; raidTags["raidpet"..i]=1; raidTags["raidpet"..i.."target"]=1; end

-- runtime data
local frame = oldframe or CreateFrame("Frame", nil, UIParent)
local tagTracked = {}
local tagID = {}
local tagListeners = {}
local logCacheUpdates = oldcacheupdates or 0
local deltaSize = 0
local delta = {}
local deltaWrapper = setmetatable({}, {
	__index = delta,
	__newindex = function() error("cannot modify shared table") end,
	__metatable = false}
)
local queue = {}


--[[ INTERNAL FUNCTIONS ]]--

local function update_events()
	local tagTracked,pairs,frame = tagTracked,pairs,frame
	-- identify necessary events
	local p_t_c = (tagTracked.target)
	local p_f_c = (tagTracked.focus)
	local u_p = (tagTracked.vehicle or tagTracked.vehicletarget or tagTracked.pet)
	if (not u_p) then
		for parent,tag in pairs(tagPet) do
			if (tagTracked[tag]) then
				u_p = true
				break
			end
		end
	end
	local u_t = (tagTracked.targettarget or tagTracked.focustarget or tagTracked.pettarget)
	if (not u_t) then
		for parent,tag in pairs(tagTarget) do
			if (tagTracked[tag]) then
				u_t = true
				break
			end
		end
	end
	local p_m_c = false
	for tag,_ in pairs(partyTags) do
		if (tagTracked[tag]) then
			p_m_c = true
			break
		end
	end
	local r_r_u = false
	for tag,_ in pairs(raidTags) do
		if (tagTracked[tag]) then
			r_r_u = true
			break
		end
	end
	-- register necessary events
	frame:UnregisterAllEvents()
	if (p_t_c) then frame:RegisterEvent("PLAYER_TARGET_CHANGED") end
	if (p_f_c) then frame:RegisterEvent("PLAYER_FOCUS_CHANGED") end
	if (u_p) then frame:RegisterEvent("UNIT_PET") end
	if (u_t) then frame:RegisterEvent("UNIT_TARGET") end
	if (p_m_c) then frame:RegisterEvent("PARTY_MEMBERS_CHANGED") end
	if (r_r_u) then frame:RegisterEvent("RAID_ROSTER_UPDATE") end
end -- update_events()

local function add_tag_listener(tag, cb, handler)
	if (not tagSupported[tag]) then
		return false
	elseif (tagListeners[tag] and tagListeners[tag][cb]) then
		return true
	elseif (not tagListeners[tag]) then
		-- establish tag's and parents' identities
		tagTracked[tag] = true
		tagListeners[tag] = {}
		tagID[tag] = API_UnitGUID(tag)
		local parent = tagParent[tag]
		while (parent) do
			tagTracked[parent] = true
			tagID[parent] = tagID[parent] or API_UnitGUID(parent)
			parent = tagParent[parent]
		end
		update_events()
	end
	tagListeners[tag][cb] = handler or false
	return true
end -- add_listener()

local update_tag_watch
update_tag_watch = function(tag)
	-- recurse to child tags first, since one of them might get cleaned up even if we still have a listener
	-- don't need player=>vehicle because we'll never call this on player anyway
	local pet_watched = (tagPet[tag] and update_tag_watch(tagPet[tag]))
	local target_watched = (tagTarget[tag] and update_tag_watch(tagTarget[tag]))
	-- but if nobody's listening, stop watching
	if (not (tagListeners[tag] or pet_watched or target_watched)) then
		tagTracked[tag] = false
		tagID[tag] = nil
		return false
	end
	return true
end -- update_tag_watch()

local function remove_tag_listener(tag, cb)
	-- make sure cb is registered
	if (not tagListeners[tag] or tagListeners[tag][cb] == nil) then
		return true
	end
	tagListeners[tag][cb] = nil
	-- if this was the tag's last listener, clean up
	if (not next(tagListeners[tag])) then
		tagListeners[tag] = nil
		local root = tag
		while (tagParent[root]) do
			root = tagParent[root]
		end
		update_tag_watch(root)
		update_events()
	end
	return true
end -- remove_listener()

local update_unit
update_unit = function(tag)
	if (not tagTracked[tag]) then return end
	local newID = API_UnitGUID(tag)
	if (newID == tagID[tag]) then return end
	-- record the change
	tagID[tag] = newID
	deltaSize = deltaSize + 1
	delta[deltaSize] = tag
	-- cascade to child tags (maximizing tail calls)
	-- don't need player=>vehicle because we'll never call this on player anyway
	if (tagPet[tag]) then
		if (tagTarget[tag]) then
			update_unit(tagPet[tag])
			return update_unit(tagTarget[tag])
		else
			return update_unit(tagPet[tag])
		end
	elseif (tagTarget[tag]) then
		return update_unit(tagTarget[tag])
	end
end -- update_unit()


--[[ RUNTIME UPGRADE ]]--

-- undocumented, for internal use only!
function LibUnitID:__dump_for_upgrade()
	return frame,tagListeners,logCacheUpdates
end -- __dump_for_upgrade()

if (oldlisteners) then
	local ok
	local n,e = 0,0
	for tag,cbs in pairs(oldlisteners) do
		for cb,handler in pairs(cbs) do
			ok = add_tag_listener(tag, cb, handler)
			if (ok) then
				n = n + 1
			else
				e = e + 1
				print("LibUnitID ERROR: Failed to re-register callback for unit '"..tag.."' during library upgrade!")
			end
		end
		wipe(cbs)
	end
	wipe(oldlisteners)
	oldlisteners = nil
--[===[@debug@
	print("LibUnitID: Upgraded from revision "..oldminor.."; "..n.." callbacks re-registered, "..e.." errors.")
--@end-debug@]===]
end


--[[ EVENT HANDLER ]]--

frame:SetScript("OnEvent", function(self, event, arg1)
	-- update affected unit(s)
--print(event.."("..tostring(arg1)..")")
	if (event == "PLAYER_TARGET_CHANGED") then
		update_unit("target")
	elseif (event == "PLAYER_FOCUS_CHANGED") then
		update_unit("focus")
	elseif (event == "UNIT_PET") then
		update_unit(tagPet[arg1] or false)
		if (arg1 == "player") then
			update_unit("vehicle")
		end
	elseif (event == "UNIT_TARGET") then
		update_unit(tagTarget[arg1] or false)
	elseif (event == "PARTY_MEMBERS_CHANGED") then
		local uu = update_unit
		uu("party1"); uu("party2"); uu("party3"); uu("party4"); uu("party5")
	elseif (event == "RAID_ROSTER_UPDATE") then
		local uu = update_unit
		uu("raid1");  uu("raid2");  uu("raid3");  uu("raid4");  uu("raid5")
		uu("raid6");  uu("raid7");  uu("raid8");  uu("raid9");  uu("raid10")
		uu("raid11"); uu("raid12"); uu("raid13"); uu("raid14"); uu("raid15")
		uu("raid16"); uu("raid17"); uu("raid18"); uu("raid19"); uu("raid20")
		uu("raid21"); uu("raid22"); uu("raid23"); uu("raid24"); uu("raid25")
		uu("raid26"); uu("raid27"); uu("raid28"); uu("raid29"); uu("raid20")
		uu("raid31"); uu("raid32"); uu("raid33"); uu("raid34"); uu("raid35")
		uu("raid36"); uu("raid37"); uu("raid38"); uu("raid39"); uu("raid40")
	end
	if (deltaSize > 0) then
		-- localize upvalues (careful with deltaSize, we need to change the upvalue itself at the end)
		local pairs,pcall = pairs,pcall
		local delta,tagID,tagListeners,queue,deltaWrapper = delta,tagID,tagListeners,queue,deltaWrapper
		local dS,tag = deltaSize--,nil
		-- add hash component to delta table, and find applicable callbacks
		for i = 1,dS do
			tag = delta[i]
			delta[tag] = tagID[tag]
			if (tagListeners[tag]) then
				for cb,handler in pairs(tagListeners[tag]) do
					queue[cb] = handler
				end
			end
		end
		delta[0] = dS
		logCacheUpdates = logCacheUpdates + dS
		-- run callbacks with deltaWrapper
		for cb,handler in pairs(queue) do
			if (handler) then
				pcall(cb, handler, deltaWrapper)
			else
				pcall(cb, deltaWrapper)
			end
		end
		-- clean up for next time
		wipe(queue)
		wipe(delta)
		deltaSize = 0
	end
end)


--[[ PUBLIC API METHODS ]]--

-- register unit

function LibUnitID:RegisterUnit(unitid, callback, handler)
	return add_tag_listener(unitid, callback, handler)
end -- RegisterUnit()

-- register party

function LibUnitID:RegisterParty(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("party"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterParty()

function LibUnitID:RegisterPartyTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("party"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyTargets()

function LibUnitID:RegisterPartyPets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("partypet"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyPets()

function LibUnitID:RegisterPartyPetTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("partypet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyPetTarges()

function LibUnitID:RegisterAllParty(callback, handler)
	local ok = 0
	for tag,_ in pairs(partyTags) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAllParty()

-- register raid

function LibUnitID:RegisterRaid(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raid"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaid()

function LibUnitID:RegisterRaidTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raid"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidTargets()

function LibUnitID:RegisterRaidPets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raidpet"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidPets()

function LibUnitID:RegisterRaidPetTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raidpet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidPetTargets()

function LibUnitID:RegisterAllRaid(callback, handler)
	local ok = 0
	for tag,_ in pairs(raidTags) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAllRaid()

-- register all

function LibUnitID:RegisterAll(callback, handler)
	local ok = 0
	for tag,_ in pairs(tagSupported) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAll()

-- unregister unit

function LibUnitID:UnregisterUnit(unitid, callback, handler)
	return remove_tag_listener(unitid, callback, handler)
end -- UnregisterUnit()

-- unregister party

function LibUnitID:UnregisterParty(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("party"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterParty()

function LibUnitID:UnregisterPartyTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("party"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyTargets()

function LibUnitID:UnregisterPartyPets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("partypet"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyPets()

function LibUnitID:UnregisterPartyPetTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("partypet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyPetTargets()

function LibUnitID:UnregisterAllParty(callback, handler)
	local ok = 0
	for tag,_ in pairs(partyTags) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAllParty()

-- unregister raid

function LibUnitID:UnregisterRaid(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raid"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaid()

function LibUnitID:UnregisterRaidTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raid"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidTargets()

function LibUnitID:UnregisterRaidPets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raidpet"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidPets()

function LibUnitID:UnregisterRaidPetTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raidpet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidPetTargets()

function LibUnitID:UnregisterAllRaid(callback, handler)
	local ok = 0
	for tag,_ in pairs(raidTags) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAllRaid()

-- unregister all

function LibUnitID:UnregisterAll(callback, handler)
	local ok = 0
	for tag,_ in pairs(tagSupported) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAll()

-- guid

function LibUnitID:GetUnitGUID(unitid)
	return tagID[unitid]
end -- GetUnitGUID()

function LibUnitID.UnitGUID(unitid) -- NOT A METHOD!  that's a dot, not a colon, Jim!
	return tagID[unitid]
end -- UnitGUID()

-- statistics

function LibUnitID:GetStatistics()
	return logCacheUpdates
end -- GetStatistics()

function LibUnitID:ResetStatistics()
	logCacheUpdates = 0
end -- ResetStatistics()

